/* mndcombi.cpp  by Wolf Jung (C) 2007-2015.  Defines classes:
   mndAngle, mndCombi, mndPath

   These classes are part of Mandel 5.13, which is free software; you can
   redistribute and / or modify them under the terms of the GNU General
   Public License as published by the Free Software Foundation; either
   version 3, or (at your option) any later version. In short: there is
   no warranty of any kind; you must redistribute the source code as well.
*/

#include <cmath>
#include <new>
#include "mndynamo.h"

int mndAngle::normalize(qulonglong &n, qulonglong &d, int &k)
{  if (!d) return 0;
   n %= d; while (!(n & 1) && !(d & 1)) { n >>= 1; d >>= 1; }
   int p; qulonglong n0, n1 = n, d1 = d, np;
   k = 0; while (!(d1 & 1)) { k++; d1 >>= 1; if (n1 >= d1) n1 -= d1; }
   n0 = n1;
   for (p = 1; p <= 65 - k; p++) { twice(n1, d1); if (n1 == n0) break; }
   if (k + p > 64) return 0;
   np = 1LL << (p - 1); np--; np <<= 1; np++; //2^p - 1 for p <= 64
   n0 = np; d >>= k; n1 = d; if (n1 > n0) { n1 = n0; n0 = d; }
   while (1) { d1 = n0 % n1; if (!d1) break; n0 = n1; n1 = d1; } //gcd n1
   n /= d/n1; n *= np/n1; d = np << k;
   return p;
}

void mndAngle::twice(qulonglong &n, qulonglong &d)
{  if (n >= d) return;
   if (!(d & 1)) { d >>= 1; if (n >= d) n -= d; return; }
   qulonglong large = 1LL; large <<= 63; //avoid overflow:
   if (n < large) { n <<= 1; if (n >= d) n -= d; return; }
   n -= large; n <<= 1; large -= (d - large); n += large;
}

/*int mndAngle::conjugate32(qulonglong &n, qulonglong &d)
{  int j, k, p = normalize(n, d, k); if (!p || p > 32 || k) return 0;
   if (p == 2) n = 3LL - n; if (p <= 2) return 1;
   qulonglong n1, d1 = d << 1, np, na = n, nb = n + d;
   if (n & 1LL) n1 = na; else n1 = nb; //A is na/d1 < ... < nb/d1
   //compute preimages of n1/d1 on the same side as forward orbit
   for (j = 2; j <= p; j++)
   {  d1 <<= 1; na <<= 1; nb <<= 1;
      np = n; for (k = j; k < p; k++) twice(np, d); np <<= j;
      if (na < np && np < nb) { if (n1 < na || nb < n1) n1 += (d1 >> 1); }
      else { if (na < n1 && n1 < nb) n1 += (d1 >> 1); }
   }
   np = n; n1 = (n1 - n)/d; n = n1;
   for (j = 1; j <= p; j++) { twice(n1, d); if (n1 == np) break; }
   if (j >= p) return p;
   while (1) { k = p % j; if (!k) break; p = j; j = k; } //gcd j
   return j;
}//conjugate32*/

int mndAngle::conjugate(qulonglong &n, qulonglong &d)
{  int j, k, p = normalize(n, d, k); if (!p || k) return 0;
   if (p == 2) n = 3LL - n; if (p <= 2) return 1;
   qulonglong np, u, na, nb, n1; //A is na < np < nb
   if (n & 1LL) { u = 0LL; na = n >> 1; np = na + (d >> 1) + 1LL; nb = np; }
   else { u = 1LL; np = n >> 1; na = np; nb = np + (d >> 1) + 1LL; }
   //Compute preimages of  n/d , such that the preperiodic angle
   // (u + n/d)/2^j  is on the same side  A  or  B  as the periodic  np/d .
   for (j = 2; j <= p; j++)
   {  if (np & 1LL) { np >>= 1; np += d >> 1; np++; } else np >>= 1;
      //Check if the old preperiodic angle is  < n/d , label this by k = 1.
      //Only for p <= 32 :  if (d*u < ((1LL << (j - 1)) - 1LL)*n)
      k = 0; n1 = n >> (p + 1 - j); //first j - 1 digits of p digits
      if (u < n1) k = 1;
      else if (u == n1)
      {  n1 |= (((1LL << (p + 1 - j)) - 1LL) & n) << (j - 1); //swap words
         if (n < n1) k = 1;
      }
      if (k)  { if (na < np && np < nb) u += 1LL << (j - 1); }
      else  { if (np <= na || nb <= np) u += 1LL << (j - 1); }
   }
   n = u; //Preperiod digits of the preimage determine the conjugate angle.
   for (j = 1; j <= p; j++) { twice(np, d); if (np == u) break; }
   if (j >= p) return p;
   while (1) { k = p % j; if (!k) break; p = j; j = k; }
   return j; //gcd of former j and p is period of characteristic point.
}//conjugate64

qulonglong mndAngle::wake(int k, int r, qulonglong &n)
{  if (k <= 0 || k >= r || r > 64) return 0LL;
   qulonglong d = 1LL; int j, s = 0; n = 1LL;
   for (j = 1; j < r; j++)
   {  s -= k; if (s < 0) s += r; if (!s) return 0LL;
      if (s > r - k) n += d << j;
   }
   d <<= (r - 1); d--; d <<= 1; d++; //2^r - 1 for r <= 64
   return d;
}

ldouble mndAngle::radians(qulonglong n, qulonglong d)
{  if (n >= d) return 0.0;
   return 2.0L * PI * ((ldouble)(n)) / ((ldouble)(d));
}

int mndAngle::setAngle(qulonglong n, qulonglong d, int &k)
{  if (n >= d) return 0;
   N = n; D = d; P = normalize(N, D, K);
   k = K; return P;
}

/*
int mndAngle::realSpider(qulonglong n, qulonglong d, double &c)
{  if (n >= d) return 0;
   int j, k, p = normalize(n, d, k); if (k || !p) return 0;
//update this old implementation:
   long int num2 = denom - num, n = num; if (2*n > denom) n = denom - num;
   double z1 = 2.0*PI/double(denom), *z = new double[period+1];
   z[1] = cos(z1*n);
   for (j = 2; j <= 32; j++)
   {  if (j == 32) return -2;  //period > 30
      n *= 2; if (n > denom) n -= denom;
      if (num < n && n < num2) return j;  //z_j < z_1
      if (n == num) { period = j - 1; break; }
      //if (n == num2) from doubling
      z[j] = cos(z1*n);
   }
   int *side = new int[period]; z1 = cos(z1*0.5*num); z[period] = 0;
   for (j = 1; j < period; j++)
   { z[j] = 2.0*(z[j] - z1); if (z[j] < 0) side[j] = 1; else side[j] =0; }
   for (k = 1; k <= 50; k++)
   {  z1 = z[1];
      for (j = 1; j < period; j++)
      {  z[j] = z[j+1] - z1; if (z[j] < 0) z[j] = 0;
         z[j] = sqrt(z[j]); if (side[j]) z[j] = -z[j];
      }
   }
   c = z[1]; delete[] z; delete[] side; return 0;
   return 0;
}//*/

double mndAngle::lambda(qulonglong N, qulonglong D,
   double eps, int nmax) // = 1.0e-8, 100
{  eps *= 0.1; if (!N || N >= D) return 1.0;
   int j, k = 0; qulonglong s = 1ULL, N1 = N >> 1, N2 = N1 + (D >> 1);
   for (j = 1; j <= 63; j++)
   {  N <<= 1; if (N >= D) N -= D;
      if (N1 < N && N <= N2) s |= (1ULL << j);
      if (!N) { k = j; break; }
   }
   if (!k) return 3.0; if (k == 1) return 2.0;
   /* Now  k  is the preperiod and  s  is the kneading sequence,
      with bit  1 ~ c1 , 2 ~ c2 , 4 ~ c3 ...
      To determine a block of the Thurston matrix,  the path  [c, beta]
      is iterated.  A[j]  encodes the path  [c' , c'']  defined in step j
      as  2^16 * c' + c'' , with  beta ~ 0  and  c ~ k .
      X[j]  is a stack of elements not yet mapped.
      F[j] = 2^16 * a' + a''  means that path j is mapped to path a'
      and path a'', with  a'' = 0  when there is only one image.
      No double array is used to encode the matrix,  because there
      are only one or two nonzero entries per column.
   */
   int c1, c2, nf, na = 1, nx = 1, Na, Nx; int *A, *F, *X;
   if (k <= 22) Na = ((k*(k + 1)) >> 1) + 2; else Na = 255; Nx = Na;
   A = new int[Na+1]; F = new int[Na+1]; X = new int[Nx+1];
   A[1] = k; X[1] = 1;
   while (nx)
   {  nf = X[nx]; c1 = A[nf]; c2 = c1 & 0xFF; c1 >>= 16; nx--;
      if ( ((s >> (k - c1)) & 1) == ((s >> (k - c2)) & 1) ) //same side
      { c2--; if (c1) c1--; c1 <<= 16; c1 |= c2; c2 = 0; }
      else //covers 0, two images
      { c2--; if (c1) c1--; c1 <<= 16; c1 |= k; c2 <<= 16; c2 |= k; }
      for (j = 1; j <= na; j++) if (A[j] == c1)
      { F[nf] = j << 16; c1 = 0; break; }
      if (c1) { na++; A[na] = c1; F[nf] = na << 16; nx++; X[nx] = na; }
      if (na >= Na - 1) return (double)(na); //or increase Na
      if (!c2) continue;
      for (j = 1; j <= na; j++) if (A[j] == c2)
      { F[nf] |= j; c2 = 0; break; }
      if (c2) { na++; A[na] = c2; F[nf] |= na; nx++; X[nx] = na; }
      if (nx == Nx || na >= Na - 1) return (double)(nx); //or increase Na , Nx
   }
   delete A; delete X;
   /* Since the block is primitve,  the largest eigenvalue is obtained
      fast by iterating the matrix and computing the growth factor.
      In principle,  repeated squaring would be faster,  but it
      requires two full double arrays and more arithmetic operations.
   */
   qulonglong *V, *Vold, *Vtemp; double lambda = 42.0, lambdaold = 0.0;
   V = new qulonglong[na+1]; Vold = new qulonglong[na+1];
   Vold[0] = (qulonglong)(na); for (j = 1; j<= na; j++) Vold[j] = 1uLL;
   s = 1ULL << 63; nf = 0;
   for (k = 1; k <= nmax; k++)
   {  for (j = 0; j <= na; j++) V[j] = 0ULL;
      for (j = 1; j <= na; j++)
      {  nx = F[j] & 0xFF; if (nx) V[nx] += Vold[j];
         V[F[j] >> 16] += Vold[j];
      }
      for (j = 1; j<= na; j++) V[0] += V[j];
      if (V[0] & s) { Vold[0] >>= 2; for (j = 0; j <= na; j++) V[j] >>= 2; }
      lambda = ((double)(V[0])) / ((double)(Vold[0]));
      lambdaold -= lambda; if (lambdaold < 0) lambdaold = -lambdaold;
      if (lambdaold < eps) nf++; else nf = 0; if (nf & 4) break;
      lambdaold = lambda; Vtemp = V; V = Vold; Vold = Vtemp;
   }
   delete Vold; delete V; delete F; return lambda;
   /* nf is needed because there may be little change at the beginning.
      This happens, e.g., for  1/4  6351/524288  3067/524288 .
      Another problem is given by approximate 2-cycles, e.g., for
      87383/1048576  86431/1048576 .
   */
}

/* Assuming that  u1  and  u2  are the numerators of conjugate angles of
   period  k , and  w  is the numerator of an angle of period  q , this
   function returns the numerator of the tuned angle of period  qk .
   The optional argument  n  requests the closest larger angle of period
   n , which is used in  toAngles()  before calling  backAngles() .
*/
qulonglong mndAngle::truncatedTuning(int k, qulonglong u1, qulonglong u2,
   int q, qulonglong w, int n) // = 0
{  if (!n) n = q*k;
   /* The implementation assumes that all arguments are > 0 , the numerators
      are appropriate for periods  k  and  q , respectively,  q >= 2 , and
      k < n <= 64 . Moreover,  (q - 2)k < n <= qk . The formulas get
      involved by allowing  qk > 64 , which happens already for
      1-2-4-5-9-13-17-22-25  in  toAngles()  since  q = 3 , k = 22, n = 25 .
   */
   int i, j; qulonglong u, L, M = 0LL, N = 0LL;
   //N  is first  n  digits of the numerator of the qk-periodic tunded angle.
   //If the n-periodic angle is smaller than the qk-periodic one,  take  ++N .
   if (q*k <= 64)
   {  for (i = q - 1; i >= 0; i--) M |= (w & (1LL << i) ? u2 : u1) << i*k;
      if (n == q*k) return M;
      //Now  N  will be the first  n  digits.  Denote the last  qk - n  digits
      //by  L , then the truncated angle will be larger if  NL > LN .
      L = 1LL << (q*k - n); L--; L &= M; N = M >> (q*k - n);
      if (M > ((L << n) | N)) return N; else return ++N;
   }
   for (i = q - 1; i >= 0; i--)
   {  j = n + (i - q)*k; u = (w & (1LL << i) ? u2 : u1);
      if (j <= 0) { N |= u >> -j; break; }
      N |= u << j;
   }
   if (n == q*k) return N;
   for (i = q - 1; i >= 0; i--) M |= (w & (1LL << i) ? u2 : u1) << i*k;
   //Now  N  is the first  n  digits and  M  is the last 64 digits of the
   //tuned numerator.  Denote the last  qk - n  digits by  L , then the
   //truncated angle will be larger if  NL > LN .
   if (q*k <= 2*n) //L is not longer than N
   {  L = 1LL << (q*k - n - 1); L--; L <<= 1; L++; L &= M;
      u1 = N >> (2*n - q*k); //first  qk - n  digits
      if (u1 > L) return N; if (u1 < L) return ++N;
      u1 = N & ((1LL << (2*n - q*k)) - 1LL); //last  2n - qk  digits
      u2 = N >> (q*k - n); //first  2n - qk  digits
      if (u1 > u2) return N; if (u1 < u2) return ++N;
      u2 = 1LL << (q*k - n - 1); u2--; u2 <<= 1; u2++; u2 &= N; //last qk - n
      if (L > u2) return N; else return ++N;
   }
   //Finally,  L is longer than  N , so replace it with  ML , where  L
   //denotes the last  n  digits of the tuned numerator and  M  is
   //the middle  qk - 2n  digits,  then the truncated angle will be
   //larger if  NML > MLN . Note that  M  will be shorter than  N  and  L .
   L = 1LL << (n - 1); L--; L <<= 1; L++; L &= M; M = 0LL;
   for (i = q - 1; i >= 0; i--)
   {  j = i*k - n; u = (w & (1LL << i) ? u2 : u1);
      if (j <= 0) { M |= u >> -j; break; }
      M |= u << j;
   }
   M &= (1LL << (q*k - 2*n)) - 1LL;
   u1 = (M << (3*n - q*k)) | (L >> (q*k - 2*n));
   if (N > u1) return N; if (N < u1) return ++N;
   u2 = L & ((1LL << (q*k - 2*n)) - 1LL);
   if (M > u2) return N; if (M < u2) return ++N;
   if (L > N) return N; else return ++N;
}

/////////////////////////////////////////////////////////////////////

int mndCombi::phi(int n) //Euler totient for  n <= 120
{  if (n <= 0 || n > 120) return 0;
   int phi = 1;  if (!(n % 2)) n /= 2; while (!(n % 2)) { n /= 2; phi *= 2; }
   if (!(n % 3)) { n /= 3; phi *= 2; } while (!(n % 3)) { n /= 3; phi *= 3; }
   if (!(n % 5)) { n /= 5; phi *= 4; } while (!(n % 5)) { n /= 5; phi *= 5; }
   if (!(n % 7)) { n /= 7; phi *= 6; } while (!(n % 7)) { n /= 7; phi *= 7; }
   if (n > 1) phi *= (n - 1); //last  n  was prime if  1 < n < 121
   return phi;
}

/* rho(m)  is the smallest index  j > m , such that the  j-th  entry of the
   kneading sequence differs from the periodic continuation of the first
   m  entries,  or simpler,  such that it differs from the entry at  j - m .
   This implementation treats the kneading sequence as *-periodic,  thus
   rho(m) <= P  for  m < P , and it does not work for  m >= P .
*/
int mndCombi::rho(int m) const
{  if (m <= 0 || m >= P) return 0;
   int j; qulonglong u = K >> m; u ^= K;
   for (j = m + 1; j < P; j++) if ((u >> (P - j)) & 1LL) return j;
   return P;
}

int mndCombi::setKneading(qulonglong k)
{  qulonglong u = 1LL; if (k <= u) { P = 0; return 0; }
   K = k; for (P = 1; P <= 64; P++) { k >>= 1; if (!k) break; }
   int m = 1, j = 0; A = u;
   while (m < P) { j = m; m = rho(m); A |= u << (m - 1); }
   //K  is the upper sequence,  if it differs from  k  in the last entry:
   k = K >> j; if (!(k & u)) K |= u; else if (K & u) K--;
   return P;
}

int mndCombi::setAddress(qulonglong a)
{  K = 1LL; if (a <= K || !(a & 1LL)) { P = 0; return 0; }
   A = a; P = 1; int i, j, p;
   for (i = 2; i <= 64; i++)
   {  a >>= 1; if (!(a & 1LL)) continue;
      p = P; P = i; K <<= P - p;
      for (j = 1; j <= (P - 1)/p; j++) K |= K >> p;
      if (K & 1LL) K--; else K++;
   }
   return P;
}

int mndCombi::fromAngle(qulonglong n, qulonglong d)
{  K = 1LL; int i; P = mndAngle::normalize(n, d, i);
   if (i || !P) { P = 0; return 0; }
   qulonglong na = n >> 1, nb = na + (d >> 1) + 1; //na < n < nb is A
   for (i = 1; i < P; i++)
   { K <<= 1; mndAngle::twice(n, d); if (na < n && n < nb) K |= 1LL; }
   setKneading(K); return P; //resets last entry to upper sequence
}

int mndCombi::renorm() const
{  int m, r; if (!P) return 0;
   for (r = 2; r < P; r++)
   {  if (!((A >> (r - 1)) & 1LL)) continue;
      for (m = r + 1; m <= P; m++)
      {  if (!((A >> (m - 1)) & 1LL)) continue;
         if (m % r) { m = 0; break; }
      }
      if (m) return r;
   }
   return 1;
}

int mndCombi::failsBS() const
{  int m, r; if (!P) return -1;
   for (m = rho(1) + 1; m < P; m++)
   {  if ((A >> (m - 1)) & 1LL) continue; //m  in address,  thus not evil
      for (r = 2; r < m; r++) if (!(m % r) && rho(r) > m) { r = 0; break; }
      if (!r) continue; //not exact period  m
      r = rho(m) % m; if (!r) return m; // ~ r = m , so  m  is in  orb(r)
      while (r < P) { if (r == m) return m; r = rho(r); }
   }
   return 0;
}//*/

/*int mndCombi::failsBS() const
{  int m, r; if (!P) return -1;
   for (m = 3; m < P; m++)
   {  if ((A >> (m - 1)) & 1LL) continue; //m  in address,  thus not evil
      for (r = 1; r < m; r++) if (!(m % r) && rho(r) > m) { r = 0; break; }
      if (!r) continue; //not exact period  m
      r = rho(m) % m; if (!r) return m; //~ r = m , so  m  is in  orb(r)
      while (r < P) { if (r == m) return m; r = rho(r); }
   }
   return 0;
}//*/

int mndCombi::count() const //assumes the combinatorics is realized
{  int j, k = 1, n, q, N = 1; if (!P) return -1; //invalid
   //Determine the denominators for all periods in the internal address:
   for (n = 2; n <= P; n++)
   {  if (!((A >> (n - 1)) & 1LL)) continue; //not in the internal address
      q = n/k; j = n % k;
      if (j) { q += 2; while (j < k) { j = rho(j); if (j == k) q--; } }
      N *= phi(q); k = n;
   }
   return N;
}

int mndCombi::toAngles(qulonglong &n1, qulonglong &n2, qulonglong &d) const
{  if (!P) return 1000; //invalid
   int j, k = 1, n, q; n1 = 0LL; n2 = 1LL; d = 1LL;
   //Determine characteristic angles for all periods in the internal address:
   for (n = 2; n <= P; n++)
   {  if (!((A >> (n - 1)) & 1LL)) continue; //not in the internal address
      //Determine the denominator of the sublimb  1/q :
      q = n/k; j = n % k;
      if (j) { q += 2; while (j < k) { j = rho(j); if (j == k) q--; } }
      //Restrict the old characteristic wake to  n-periodic angles:
      d = n1; n1 = mndAngle::truncatedTuning(k, d, n2, q, 1LL, n);
      n2 = mndAngle::truncatedTuning(k, d, n2, q, 2LL, n);
      d = 1LL << (n - 1); d--; d <<= 1; d++; //2^n - 1 for n <= 64
      if (!(n % k)) { k = n; continue; } //satellite bifurcation
      k = n; n2--; // n1/d  is larger and  n2/d  is smaller than  qk-angles.
      if (n2 - n1 == 1LL) continue; //always admissible
#ifdef COUNT_EARLY_RETURNS
      qulonglong u, u1 = n1, u2 = n2;
#endif
      //Iterate the old wake backwards along the new kneading sequence:
      if (backAngles(n, n1, n2, d)) return n; //Error, thus not admissible
#ifdef COUNT_EARLY_RETURNS
      if (n == P) //count the returns to previous wake
      {  q = 0; u = n1;
         for (j = 1; j < n; j++)
         { mndAngle::twice(u, d); if (u1 <= u && u <= u2) q++; }
         return -q;
      }
#endif
   }
   return 0;
}

int mndCombi::backAngles(int n,
   qulonglong &n1, qulonglong &n2, qulonglong d) const
{  int j; qulonglong N1, N2, N3, N4;
   //The preimages  [N1, N2]  and  [N3, N4]  of the reduced characteristic
   //wake  [n1, n2]  will be pulled back following the kneading sequence.
   N1 = (n1 + 1LL) >> 1; N2 = n2 >> 1;
   N4 = (d >> 1) + 1LL; N3 = (n1 >> 1) + N4; N4 += (n2 - 1LL) >> 1;
   for (j = n - 1; j > 0; j--)
   {  //When preimages belong to the characteristic wake,  try subdivision:
      if (n1 <= N1 && N1 <= n2)
      {  if (N1 > N4) { N1 = N3; N4 = N2; }
         N2 = n2; n2 = N1 - 1LL; //first subinterval
         if (n1 < n2 && !backAngles(n, n1, n2, d)) return 0;
         n1 = N4 + 1LL; n2 = N2; //second subinterval
         if (n1 < n2) return backAngles(n, n1, n2, d); else return 1;
      }
      //The preimages of the intervals are shrunk to  1/2  at least,
      //or more depending on a numerator being even or odd:
      if ((K >> (P - j)) & 1LL) //A in kneading sequence
      {  if (N1 > n2) { N1++; N1 >>= 1; N2 >>= 1; }
         else { N1 >>= 1; N1++; N1 += d >> 1; N2++; N2 >>= 1; N2 += d >> 1; }
         if (N3 > n2) { N3++; N3 >>= 1; N4 >>= 1; }
         else { N3 >>= 1; N3++; N3 += d >> 1; N4++; N4 >>= 1; N4 += d >> 1; }
      }
      else //B in kneading sequence
      {  if (N2 < n1) { N1++; N1 >>= 1; N2 >>= 1; }
         else { N1 >>= 1; N1++; N1 += d >> 1; N2++; N2 >>= 1; N2 += d >> 1; }
         if (N4 < n1) { N3++; N3 >>= 1; N4 >>= 1; }
         else { N3 >>= 1; N3++; N3 += d >> 1; N4++; N4 >>= 1; N4 += d >> 1; }
      }
      if (N1 > N2 || N3 > N4) return 1;
   }
   //Now  N1 = N2  and  N3 = N4  are the new characteristic angles times d.
   if (N1 > N3) { n2 = N1; n1 = N3; } else { n1 = N1; n2 = N3; }
   return 0;
}//backAngles

/* Computing external angles from an internal address or kneading sequence

Suppose the external angles of  1-...-k  are known already and those of
1-...-k-n  are to be determined.  First compute the denominator  q  from
the Bruin-Schleicher formula:  q = n/k  when  k | n  and
q = [n/k] + 1  or  [n/k] + 2  if  n in orb_rho(n % k)  or not,  respectively.
The implementation is choosing the  1/q-sublimb,  but it could be extended
to work with any angled internal address.

The angles  t1  and  t2  of the 1/q-sublimb are found with Douady tuning.
If  n = qk , we are done.  If not,  this  qk-periodic sector is iterated
back along the kneading sequence,  giving a sequence of strips around the
desired  n-periodic rays.  The strip around  z = 0  is defined by the
intervals  [T1, T2]  and  [T3, T4] . Its complementary sectors are contained
in the pieces A and B.  So for  j = n-1 ... j = 1 , iterating back means:
* If the  j-th  entry is  A , then  T1  and  T2  are mapped with
  T -> T/2  if  T1 > t2  and with  T -> (T + 1)/2  if  T2 < t1 .
* If the  j-th  entry is  B , it is vice versa.
* [T3, T4]  is iterated back with the same formulas.
It may happen that a preimage of the strip is contained in the sector,  so
t1 < T1 < T2 < T3 < T4 < t2  (after interchanging  Ti  if necessary). Then
we do not know which of its two preimages will be on the correct side,
and the algorithm is branching to try both possibilities:
* Assuming the characteristic angles of period  n  are in  (t1, T1) ,
  the preimages in  (0, 1/2)  will be on side  A  and those in  (1/2, 1)
  will be on side  B .
* Assuming the characteristic angles of period  n  are in  (T4, t2) ,  the
  other preimage of the strip will be the correct one.

Each branch turns up a narrow strip at  j = 1 , which should contain the
desired characteristic rays of period  n . These things can go wrong:
*  [T1, T2]  and  [T3, T4]  do not belong to the reduced wake assumed.
*  [T1, T2]  and  [T3, T4]  need not contain  n-periodic angles.
The second error happens only when the algorithm was restarted at  j = n
with the reduced wake,  and then the branch is stopped before the
first error will happen.

Proposition:
1) If nothing goes wrong in a branch of this algorithm,  it will give the
   correct angles.  In particular, this branch is the only one going through,
   and the internal address is admissible.
2) If all branches are failing,  the internal address will not be admissible.

Proof:
1) The angles will have the desired kneading sequence and belong to the
   1/q-sublimb of period k .
2) If the address is admissible,  consider the dynamics for the corresponding
   center.  There is a strip around the critical value,  which is iterated
   2:1  to the characteristic wake in  n  iterations.  The corresponding
   image strips are contained in larger strips,  which are preimages of the
   characteristic wake of period  qk : iterating it backwards is well-defined,
   since otherwise a large strip would have bifurcated somewhere between
   the centers of periods  qk  and  n . This does not happen for preimages
   of order  <= n  of the strip defined by the precharacteristic points,
   since these do not contain  0 , because there is a stable branch point
   separating the critical value from the other marked points in the  qk-wake.
   So,  if the algorithm is branching,  there will be a branch assuming the
   correct location of the  n-wake between the preimages of the  qk-wake,
   and all of the strips are well-defined strips around the  n-periodic rays.
For the other branches,  or if the address is not admissible,  we may still
assume that the intervals  [T1, T2]  and  [T3, T4]  are describing unlinked
strips,  because this will be the case for the center of period  qk .
The implementation is simplified by restarting the iteration with the
smaller intervals  (t1, T1)  or  (T4, t2)  when the algorithm is branching.
Considering these subsets is possible although they are not wakes.

Question:
Can the critical orbit return early more than once for an admissible
combinatorics?  If not, the recursive subdivision could be simplified.
There are only single early returns for periods  n <= 33 , and the first
example is given by  1-2-6-7-13-14 .

Implementation with  n-periodic angles:
t1  and  t2  are replaced with  n-periodic angles  n1/d > t1  and  n2/d < t2 .
Choosing the closest possible angles is quite involved,  since for large  qk
we cannot take  n2 = [d*t2]  because floating-point arithmetics is inaccurate.
The rest of the algorithm is simplified for  n <= 64, because the denominators
are not increased under iterating backwards (half-integer numerators are
rounded to the appropriate side),  and it is easy to see if the strip does
not contain an  n-periodic angle.  While a preimage of the  qk-periodic
sector would never touch its boundary,  ni/d  may have a period dividing  n .
Examples:  for the non-admissible  1-2-4-5-6 , both  n1/d  and  n2/d  are
3-periodic instead of  6-periodic.  For the admissible  1-3-6-7-10-12 ,
n1/d  is  12-periodic but  n2/d  is  4-periodic.
*/

//////////////////////////////////////////////////////////////////////////

int mandelbrotPath::capture(int k0, int p0, mandelPathinfo *mpi)
{  type = 1; k = k0; p = p0; rp = mpi->rc[0]; ip = mpi->ic[0];
   int i, m; double x, y, w;
   while (mpi->n < M)
   {  double *rz = new double[2*mpi->n + 1], *iz = new double[2*mpi->n + 1];
      for (m = 0; m <= mpi->n; m++) { rz[2*m] = mpi->rc[m]; iz[2*m] = mpi->ic[m]; }
      for (m = 1; m <= mpi->n; m++)
      {  rz[2*m - 1] = 0.5*(mpi->rc[m - 1] + mpi->rc[m]);
	 iz[2*m - 1] = 0.5*(mpi->ic[m - 1] + mpi->ic[m]);
      }
      mpi->n *= 2;
      delete[] mpi->rc; delete mpi->ic; mpi->rc = rz; mpi->ic = iz;
   }
   delete[] X; delete[] Y;
   X = new double[1 + M*(k+p)]; Y = new double[1 + M*(k+p)];
   for (m = M - 1; m >= 0; m--)
   {  X[1 + m*(k+p)] = mpi->rc[mpi->n - (mpi->n*m)/M];
      Y[1 + m*(k+p)] = mpi->ic[mpi->n - (mpi->n*m)/M];
   }
   x = X[1]; y = Y[1];
   for (i = 2; i < k+p; i++)
   {  w = x*x - y*y + rp; y = 2.0*x*y + ip; x = w;
      if (x*x + y*y > 4.0) return -1;
      X[i] = x; Y[i] = y;
   }
   if (k) { X[k+p] = -X[k]; Y[k+p] = -Y[k]; }
   else { X[p] = 0.0; Y[p] = 0.0; } //and iterate backwards!!!!!!!!
   for (m = M - 1; m > 0; m--) for (i = 2; i <= k+p; i++)
   { X[i+m*(k+p)] = X[i]; Y[i+m*(k+p)] = Y[i]; }
   if (draw)
   {  delete[] X0; delete[] Y0;
      X0 = new double[1 + M*(k+p)]; Y0 = new double[1 + M*(k+p)];
      for (i = 1; i <= M*(k+p); i++) { X0[i] = X[i]; Y0[i] = Y[i]; }
   }
   return 0;
}

int mandelbrotPath::spider(qulonglong N, qulonglong D, double logR1, int)
{  type = 2;
  
   return 0;
}

int mandelbrotPath::step(double &rc, double &ic)
{  if (!type) return -1;
   int i, oldset = 0, offset = (M - 1)*(k+p); double u, v, w, x, y;
   //for (m = M - 1; m >= 0; m--)
   while (offset >= 0)
   {  u = X[1 + offset]; v = Y[1 + offset];
      for (i = 1; i < k+p; i++)
      {  mndynamics::root(X[i + 1 + offset] - u, Y[i + 1 + offset] - v, x, y);
	 if (x*X[i + oldset] + y*Y[i + oldset] < 0.0) { x = -x; y = -y; }
	 X[i + offset] = x; Y[i + offset] = y;
      }
      if (k) //else z_p stays 0
      { X[k+p + offset] = -X[k + offset]; Y[k+p + offset] = -Y[k + offset]; }
      oldset = offset; offset -= k+p;
   }
   rc = X[1]; ic = Y[1]; return 0;
}

int mandelbrotPath::sequence(int n, int f, int fps, mandelPathinfo *mpi)
{  if (!type || !draw) return -1;
   int i, m = (fps - f)*M/fps; double x, y;
   delete[] mpi->coeff; delete[] mpi->rc; delete[] mpi->ic; mpi->coeff = 0;
   mpi->rc = new double[n + 1]; mpi->ic = new double[n + 1]; mpi->n = n;
   mpi->rc[0] = rp; mpi->ic[0] = ip;
   for (i = 1; i <= M*(k+p); i++) { X[i] = X0[i]; Y[i] = Y0[i]; }
   mpi->rc[1] = X[1+m*(k+p)]; mpi->ic[1] = Y[1+m*(k+p)];
   for (i = 2; i <= n; i++)
   {  if (step(x, y)) return 1;
      mpi->rc[i] = X[1+m*(k+p)]; mpi->ic[i] = Y[1+m*(k+p)];
   }
   return 0;
}

//////////////////////////////////////////////////////////////////////////

int mandelPathTwo::capture(int k0, int p0, mandelPathinfo *mpi)
{  type = 1; k = k0; p = p0; rp = -mpi->rc[0]; ip = -mpi->ic[0];
   int i, m; double x, y, u, v, w;
   while (mpi->n < M)
   {  double *rz = new double[2*mpi->n + 1], *iz = new double[2*mpi->n + 1];
      for (m = 0; m <= mpi->n; m++) { rz[2*m] = mpi->rc[m]; iz[2*m] = mpi->ic[m]; }
      for (m = 1; m <= mpi->n; m++)
      {  rz[2*m - 1] = 0.5*(mpi->rc[m - 1] + mpi->rc[m]);
         iz[2*m - 1] = 0.5*(mpi->ic[m - 1] + mpi->ic[m]);
      }
      mpi->n *= 2;
      delete[] mpi->rc; delete mpi->ic; mpi->rc = rz; mpi->ic = iz;
   }
   delete[] X; delete[] Y;
   X = new double[1 + M*(k+p)]; Y = new double[1 + M*(k+p)];
   for (m = M - 1; m >= 0; m--)
   {  X[1 + m*(k+p)] = mpi->rc[mpi->n - (mpi->n*m)/M];
      Y[1 + m*(k+p)] = mpi->ic[mpi->n - (mpi->n*m)/M];
   }
   x = X[1]; y = Y[1];
   for (i = 2; i < k+p; i++)
   {  u = rp + 1.0, v = ip, w = x*x - y*y - 1.0; y = 2.0*x*y; x = w;
      w = x*x + y*y; if (w < 1e-100) { type = 0; return -1; } x /= w; y /= w;
      w = u*x + v*y; y = v*x - u*y; x = w + 1.0; X[i] = x; Y[i] = y;
   }
   if (k)
   {  if (p) { X[k+p] = -X[k]; Y[k+p] = -Y[k]; }
      else { X[k] = -1.0; Y[k] = 0.0; } //and iterate backwards !!!!!!!!!!!
   }
   else { X[p] = 0.0; Y[p] = 0.0; } //and iterate backwards !!!!!!!!!!!
   for (m = M - 1; m > 0; m--) for (i = 2; i <= k+p; i++)
   { X[i+m*(k+p)] = X[i]; Y[i+m*(k+p)] = Y[i]; }
   if (draw)
   {  delete[] X0; delete[] Y0;
      X0 = new double[1 + M*(k+p)]; Y0 = new double[1 + M*(k+p)];
      for (i = 1; i <= M*(k+p); i++) { X0[i] = X[i]; Y0[i] = Y[i]; }
   }
   return 0;
}

int mandelPathTwo::spider(qulonglong N, qulonglong D, double logR1, int)
{  type = 2;
  
   return 0;
}

int mandelPathTwo::mating(int k0, int p0, double rc, double ic,
   double logR1, int)
{  type = 2; k = k0; p = p0; rp = rc; ip = ic; twoLogR = 2.0*logR1;
   int i, m; double x = rp, y = ip, u, v, w, RS, omt;
   delete[] X; delete[] Y;
   X = new double[1 + M*(k+p)]; Y = new double[1 + M*(k+p)]; X[1] = x; Y[1] = y;
   for (i = 2; i < k+p; i++)
   {  w = x*x - y*y + rp; y = 2.0*x*y + ip; x = w; X[i] = x; Y[i] = y;
      if (x*x + y*y > 4.0) { type = 0; return -1; }
   }
   if (k) { X[k+p] = -X[k]; Y[k+p] = -Y[k]; } else { X[p] = 0.0; Y[p] = 0.0; }
   for (m = M - 1; m >= 0; m--)
   {  omt = ((double)(m))/((double)(M)); RS = exp(twoLogR*exp(logTwo*omt));
      for (i = 1; i <= k+p; i++)
      {  u = omt*(X[i] - rp) - RS; v = omt*(ip - Y[i]);
         w = u*u + v*v; u /= w; v /= w;
         w = u*X[i] - v*Y[i]; Y[i+m*(k+p)] = u*Y[i] + v*X[i]; X[i+m*(k+p)] = w;
      }
   }
   if (draw)
   {  delete[] X0; delete[] Y0;
      X0 = new double[1 + M*(k+p)]; Y0 = new double[1 + M*(k+p)];
      for (i = 1; i <= M*(k+p); i++) { X0[i] = X[i]; Y0[i] = Y[i]; }
   }
   return 0;
}

int mandelPathTwo::anti(int k0, int p0, double rQ, double iQ,
   double logR1, int)
{  type = 3; k = k0; p = p0; rp = rQ; ip = iQ; twoLogR = 2.0*logR1;
   int i, m; double x = -rp, y = -ip, u, v, w, RS,
      omt = rp*rp + ip*ip, ri = rp/omt, ii = -ip/omt;
   delete[] X; delete[] Y;
   X = new double[1 + M*(k+p)]; Y = new double[1 + M*(k+p)]; X[1] = x; Y[1] = y;
   for (i = 2; i < k+p; i++)
   {  u = x*x - y*y; v = 2.0*x*y;
      if (i & 1) { u--; x = rp*u - ip*v; y = rp*v + ip*u; }
      else { x = ri*u - ii*v + 1.0; y = ri*v + ii*u; }
      X[i] = x; Y[i] = y;
      if (x*x + y*y > 16.0) { type = 0; return -1; }
      if (x*x + y*y < 0.1 && (i & 1) ) { k = i - 1; p = 0; }
   }
   if (k)
   {  if (p) { X[k+p] = -X[k]; Y[k+p] = -Y[k]; }
      else { X[k] = -1.0; Y[k] = 0.0; }
   }
   else { X[p] = 0.0; Y[p] = 0.0; }
   for (m = M - 1; m >= 0; m--)
   {  omt = ((double)(m))/((double)(M)); RS = exp(twoLogR*exp(logTwo*omt));
      for (i = 1; i <= k+p; i += 2)
      {  x = X[i]; y = Y[i]; w = x*x + y*y;
         if (w < 1e-100)  { type = 0; return -1; }
         x /= w; y /= -w; u = RS + omt*rp; v = omt*ip;
         X[i+m*(k+p)] = u*x - v*y + omt; Y[i+m*(k+p)] = u*y + v*x;
      }
      omt /= RS; RS = omt*ip; omt *= rp; 
      for (i = 2; i <= k+p; i += 2)
      {  x = X[i]; y = Y[i];
         u = 1.0 + omt*(1.0 - x) + RS*y; v = RS*(1.0 - x) - omt*y;
         w = u*u + v*v; if (w < 1e-100) { type = 0; return -1; }
         u /= w; v /= -w;
         X[i+m*(k+p)] = u*x - v*y; Y[i+m*(k+p)] = u*y + v*x;
      }
   }
   if (draw)
   {  delete[] X0; delete[] Y0;
      X0 = new double[1 + M*(k+p)]; Y0 = new double[1 + M*(k+p)];
      for (i = 1; i <= M*(k+p); i++) { X0[i] = X[i]; Y0[i] = Y[i]; }
   }
   return 0;
}

int mandelPathTwo::step(double &rc, double &ic)
{  if (!type) return -1;
   int i, oldset = 0, offset = (M - 1)*(k+p); double u, v, w, x, y;
   //for (m = M - 1; m >= 0; m--)
   while (offset >= 0)
   {  u = X[1 + offset] - 1.0; v = Y[1 + offset];
      for (i = 1; i < k+p; i++)
      {  x = 1.0 - X[i + 1 + offset]; y = Y[i + 1 + offset];
         w = x*x + y*y; if (w < 1e-20) return 1;
         x /= w; y /= w; mndynamics::root(1.0 + u*x - v*y, u*y + v*x, x, y);
         if (x*X[i + oldset] + y*Y[i + oldset] < 0.0) { x = -x; y = -y; }
         X[i + offset] = x; Y[i + offset] = y;
      }
      if (k) //else z_p stays 0
      {  if (p)
         { X[k+p + offset] = -X[k + offset]; Y[k+p + offset] = -Y[k + offset]; }
         else //needed for antimating,  and for capture path when  k = 1
         { X[k+p + offset] = -1.0; Y[k+p + offset] = 0.0; }
      }
      oldset = offset; offset -= k+p;
   }
   rc = -X[1]; ic = -Y[1]; return 0;
}

int mandelPathTwo::sequence(int n, int f, int fps, mandelPathinfo *mpi)
{  if (!type || !draw) return -1;
   int i, m = (fps - f)*M/fps; double x, y;
   delete[] mpi->coeff; delete[] mpi->rc; delete[] mpi->ic;
   if (type == 1) mpi->coeff = 0;
   else
   {  mpi->coeff = new double[4];
      double omt = ((double)(m))/((double)(M)),
         RS = exp(twoLogR*exp(logTwo*omt));
      mpi->coeff[0] = RS;
      if (type == 2)
      { mpi->coeff[1] = -(RS + omt*rp); mpi->coeff[2] = -omt*ip; }
      mpi->coeff[3] = omt;
   }
   mpi->rc = new double[n + 1]; mpi->ic = new double[n + 1]; mpi->n = n;
   mpi->rc[0] = rp; mpi->ic[0] = ip;
   for (i = 1; i <= M*(k+p); i++) { X[i] = X0[i]; Y[i] = Y0[i]; }
   mpi->rc[1] = -X[1+m*(k+p)]; mpi->ic[1] = -Y[1+m*(k+p)];
   for (i = 2; i <= n; i++)
   {  if (step(x, y)) return 1;
      mpi->rc[i] = -X[1+m*(k+p)]; mpi->ic[i] = -Y[1+m*(k+p)];
   }
   return 0;
}
//////////////////////////////////////////////////////////////////////////

int mandelPathThree::capture(int k0, int p0, mandelPathinfo *mpi) //!!!!!!!!!!!
{  type = 1; k = k0; p = p0; rp = -mpi->rc[0]; ip = -mpi->ic[0];
   int i, m; double x, y, u, v, w;
   while (mpi->n < M)
   {  double *rz = new double[2*mpi->n + 1], *iz = new double[2*mpi->n + 1];
      for (m = 0; m <= mpi->n; m++) { rz[2*m] = mpi->rc[m]; iz[2*m] = mpi->ic[m]; }
      for (m = 1; m <= mpi->n; m++)
      {  rz[2*m - 1] = 0.5*(mpi->rc[m - 1] + mpi->rc[m]);
         iz[2*m - 1] = 0.5*(mpi->ic[m - 1] + mpi->ic[m]);
      }
      mpi->n *= 2;
      delete[] mpi->rc; delete mpi->ic; mpi->rc = rz; mpi->ic = iz;
   }
   delete[] X; delete[] Y;
   X = new double[1 + M*(k+p)]; Y = new double[1 + M*(k+p)];
   for (m = M - 1; m >= 0; m--)
   {  X[1 + m*(k+p)] = mpi->rc[mpi->n - (mpi->n*m)/M];
      Y[1 + m*(k+p)] = mpi->ic[mpi->n - (mpi->n*m)/M];
   }
   x = X[1]; y = Y[1];
   for (i = 2; i < k+p; i++)
   {  u = rp + 1.0, v = ip, w = x*x - y*y - 1.0; y = 2.0*x*y; x = w;
      w = x*x + y*y; if (w < 1e-100) { type = 0; return -1; } x /= w; y /= w;
      w = u*x + v*y; y = v*x - u*y; x = w + 1.0; X[i] = x; Y[i] = y;
   }
   if (k)
   {  if (p) { X[k+p] = -X[k]; Y[k+p] = -Y[k]; }
      else { X[k] = -1.0; Y[k] = 0.0; } //and iterate backwards !!!!!!!!!!!
   }
   else { X[p] = 0.0; Y[p] = 0.0; } //and iterate backwards !!!!!!!!!!!
   for (m = M - 1; m > 0; m--) for (i = 2; i <= k+p; i++)
   { X[i+m*(k+p)] = X[i]; Y[i+m*(k+p)] = Y[i]; }
   if (draw)
   {  delete[] X0; delete[] Y0;
      X0 = new double[1 + M*(k+p)]; Y0 = new double[1 + M*(k+p)];
      for (i = 1; i <= M*(k+p); i++) { X0[i] = X[i]; Y0[i] = Y[i]; }
   }
   return 0;
}

int mandelPathThree::spider(qulonglong N, qulonglong D, double logR1, int which)
{  type = 2;
  
   return 0;
}
/*
q2   -1.754877666246693  +0.000000000000000
q3   -0.122561166876654  +0.744861766619744
q4   -0.122561166876654  -0.744861766619744
c2   1.324717957244746  +0.000000000000000
c3   -0.662358978622373  +0.562279512062301
c4   -0.662358978622373  -0.562279512062301
*/
int mandelPathThree::mating(int k0, int p0, double rc, double ic,
   double logR1, int which)
{  type = 2; k = k0; p = p0; rp = rc; ip = ic; twoLogR = 2.0*logR1;
   int i, m; double x = rp, y = ip, u, v, w, RS, omt;
   delete[] X; delete[] Y;
   X = new double[1 + M*(k+p)]; Y = new double[1 + M*(k+p)]; X[1] = x; Y[1] = y;
   for (i = 2; i < k+p; i++)
   {  w = x*x - y*y + rp; y = 2.0*x*y + ip; x = w; X[i] = x; Y[i] = y;
      if (x*x + y*y > 4.0) { type = 0; return -1; }
   }
   if (k) { X[k+p] = -X[k]; Y[k+p] = -Y[k]; } else { X[p] = 0.0; Y[p] = 0.0; }
   for (m = M - 1; m >= 0; m--)
   {  omt = ((double)(m))/((double)(M)); RS = exp(twoLogR*exp(logTwo*omt));
      for (i = 1; i <= k+p; i++)
      {  u = omt*(X[i] - rp) - RS; v = omt*(ip - Y[i]);
         w = u*u + v*v; u /= w; v /= w;
         w = u*X[i] - v*Y[i]; Y[i+m*(k+p)] = u*Y[i] + v*X[i]; X[i+m*(k+p)] = w;
      }
   }
   if (draw)
   {  delete[] X0; delete[] Y0;
      X0 = new double[1 + M*(k+p)]; Y0 = new double[1 + M*(k+p)];
      for (i = 1; i <= M*(k+p); i++) { X0[i] = X[i]; Y0[i] = Y[i]; }
   }
   return 0;
}

int mandelPathThree::step(double &rc, double &ic)
{  if (!type) return -1;
   int i, oldset = 0, offset = (M - 1)*(k+p); double u, v, w, x, y;
   //for (m = M - 1; m >= 0; m--)
   while (offset >= 0)
   {  u = X[1 + offset] - 1.0; v = Y[1 + offset];
      for (i = 1; i < k+p; i++)
      {  x = 1.0 - X[i + 1 + offset]; y = Y[i + 1 + offset];
         w = x*x + y*y; if (w < 1e-20) return 1;
         x /= w; y /= w; mndynamics::root(1.0 + u*x - v*y, u*y + v*x, x, y);
         if (x*X[i + oldset] + y*Y[i + oldset] < 0.0) { x = -x; y = -y; }
         X[i + offset] = x; Y[i + offset] = y;
      }
      if (k) //else z_p stays 0
      {  if (p)
         { X[k+p + offset] = -X[k + offset]; Y[k+p + offset] = -Y[k + offset]; }
         else //needed for antimating,  and for capture path when  k = 1
         { X[k+p + offset] = -1.0; Y[k+p + offset] = 0.0; }
      }
      oldset = offset; offset -= k+p;
   }
   rc = -X[1]; ic = -Y[1]; return 0;
}

int mandelPathThree::sequence(int n, int f, int fps, mandelPathinfo *mpi)
{  if (!type || !draw) return -1;
   int i, m = (fps - f)*M/fps; double x, y;
   delete[] mpi->coeff; delete[] mpi->rc; delete[] mpi->ic;
   if (type == 1) mpi->coeff = 0;
   else
   {  mpi->coeff = new double[4];
      double omt = ((double)(m))/((double)(M)),
         RS = exp(twoLogR*exp(logTwo*omt));
      mpi->coeff[0] = RS;
      if (type == 2)
      { mpi->coeff[1] = -(RS + omt*rp); mpi->coeff[2] = -omt*ip; }
      mpi->coeff[3] = omt;
   }
   mpi->rc = new double[n + 1]; mpi->ic = new double[n + 1]; mpi->n = n;
   mpi->rc[0] = rp; mpi->ic[0] = ip;
   for (i = 1; i <= M*(k+p); i++) { X[i] = X0[i]; Y[i] = Y0[i]; }
   mpi->rc[1] = -X[1+m*(k+p)]; mpi->ic[1] = -Y[1+m*(k+p)];
   for (i = 2; i <= n; i++)
   {  if (step(x, y)) return 1;
      mpi->rc[i] = -X[1+m*(k+p)]; mpi->ic[i] = -Y[1+m*(k+p)];
   }
   return 0;
}

////////////////////////////////////////////////////////////////////////